package org.richfaces.realworld.domain;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

import org.hibernate.validator.Length;
import org.hibernate.validator.NotEmpty;
import org.hibernate.validator.NotNull;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.AutoCreate;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Scope;

@Entity
@Name("shelf")
@Scope(ScopeType.CONVERSATION)
@Table(name = "shelves")
@AutoCreate
public class Shelf implements Serializable {

    private static final long serialVersionUID = -7042878411608396483L;

    @Id
    @GeneratedValue
    @Column(name = "SHELF_ID")
    private Long id = null;

    @Column(length = 255, nullable = false)
    @NotNull
    @NotEmpty
    @Length(min = 3, max = 50)
    private String name;

    @Column(length = 1024)
    private String description;

    @ManyToOne(fetch = FetchType.LAZY)
    @JoinColumn(name = "USER_ID", referencedColumnName = "USER_ID")
    private User owner;

    @OneToMany(mappedBy = "shelf", cascade = {CascadeType.ALL})
    @org.hibernate.annotations.OrderBy(clause = "NAME asc")
    private List<Album> albums = new ArrayList<Album>();

    @NotNull
    private boolean shared;

    @Temporal(TemporalType.TIMESTAMP)
    private Date created;

    /**
     * Getter for property preDefined
     *
     * @return is this shelf is predefined
     */
    public boolean isPreDefined() {
		return getOwner() != null? getOwner().isPreDefined() : false;
	}

	/**
     * Getter for property id
     *
     * @return id of album
     */
    public Long getId() {
        return id;
    }

    /**
     * Getter for property name
     *
     * @return name of album
     */
    public String getName() {
        return name;
    }

    /**
     * Setter for property name
     *
     * @param name -
     *             name of album
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Getter for property description
     *
     * @return description of album
     */
    public String getDescription() {
        return description;
    }

    /**
     * Setter for property description
     *
     * @param description -
     *                    description of album
     */
    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * Getter for property albums
     *
     * @return collection of albums, belongs to this shelf.
     */
    public List<Album> getAlbums() {
        return albums;
    }

    /**
     * Getter for property owner
     *
     * @return user, that created this shelf
     */
    public User getOwner() {
        return owner;
    }

    /**
     * Setter for property owner
     *
     * @param owner -owner of shelf
     */
    public void setOwner(User owner) {
        this.owner = owner;
    }

    /**
     * Getter for property created
     *
     * @return date of creation of this shelf
     */
    public Date getCreated() {
        return created;
    }

    /**
     * Setter for property created
     *
     * @param created - date of creation
     */
    public void setCreated(Date created) {
        this.created = created;
    }

    /**
     * Getter for property shared. If true - all users can view this shelfs and albums and images, contained in this shelf,
     * otherwise this shelf can view only owner.
     *
     * @return shared
     */
    public boolean isShared() {
        return shared;
    }

    /**
     * Setter for property shared
     *
     * @param shared - determine is this shelf will be accessible by other users.
     */
    public void setShared(boolean shared) {
        this.shared = shared;
    }
    
 // ********************** Business Methods ********************** //
    
    /**
     * @return List of unvisited images
     */
    public List<Image> getUnvisitedImages() {
    	List<Image> unvisitedImages = new ArrayList<Image>();
    	for(Album a : getAlbums()){
    		for(Image i : a.getImages()){
    			if(i.isNew()){
    				unvisitedImages.add(i);
    			}
    		}
    	}
        return unvisitedImages;
    }
    
    /**
     * @return List of images, belongs to this shelf
     */
    public List<Image> getImages() {
    	List<Image> images = new ArrayList<Image>();
    	for(Album a : getAlbums()){
    		images.addAll(a.getImages());
    	}
        return images;
    }
    
    /**
     * This method add album to collection of albums of current shelf
     *
     * @param album -
     *              album to add
     */
    public void addAlbum(Album album) {
        if (album == null)
            throw new IllegalArgumentException("Null album!");
        if (album.getShelf() != null && !album.getShelf().getAlbums().contains(this)){
            // remove from previous shelf
        	album.getShelf().removeAlbum(album);
            album.setShelf(this);
            albums.add(album);
        }	
    }

    /**
     * This method remove album from collection of albums of album
     *
     * @param album -
     *              album to remove
     */
    public void removeAlbum(Album album) {
        if (album == null)
            throw new IllegalArgumentException("Null album!");
        if(album.getShelf().equals(this)){
        	album.setShelf(null);
            albums.remove(album);
        }else{
        	throw new IllegalArgumentException("This Shelf not contain this album!");
        }
    }

    /**
     * This method return first album of current shelf or null if shelf haven't albums.
     *
     * @return first album of shelf or null
     */
    public Album getFirstAlbum() {
        if (this.albums.size() > 0) {
            return this.albums.get(0);
        }
        return null;
    }
    
    /**
     * Return relative path of this shelf in file-system(relative to uploadRoot parameter)
     *
     */
    public String getPath(){
		if(getOwner() != null && getOwner().getPath() != null){
			return getOwner().getPath() + this.getId().toString() + File.separator;
		}
		return null;
	}

    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Shelf shelf = (Shelf) o;

        if (id != null ? !id.equals(shelf.id) : shelf.id != null) return false;
        if (owner != null ? !owner.equals(shelf.owner) : shelf.owner != null) return false;
        if (!name.equals(shelf.name)) return false;

        return true;
    }

    public int hashCode() {
        int result;
        result = (id != null ? id.hashCode() : 0);
        result = 31 * result + name.hashCode();
        result = 31 * result + (owner != null ? owner.hashCode() : 0);
        return result;
    }
}