/**
 * License Agreement.
 *
 *  JBoss RichFaces - Ajax4jsf Component Library
 *
 * Copyright (C) 2007  Exadel, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */
package org.richfaces.realworld.manager;

import java.io.Serializable;
import java.security.MessageDigest;

import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.security.auth.login.LoginException;

import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.End;
import org.jboss.seam.annotations.In;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Scope;
import org.jboss.seam.annotations.web.RequestParameter;
import org.jboss.seam.contexts.Contexts;
import org.jboss.seam.core.Events;
import org.jboss.seam.faces.FacesMessages;
import org.jboss.seam.security.Credentials;
import org.jboss.seam.security.Identity;
import org.jboss.seam.util.Hex;
import org.richfaces.realworld.domain.User;
import org.richfaces.realworld.service.Constants;
import org.richfaces.realworld.service.IUserAction;

@Name("authenticator")
@Scope(ScopeType.CONVERSATION)
public class Authenticator implements Serializable{

	private static final long serialVersionUID = -4585673256547342140L;

	@RequestParameter protected Long startConversation;
	
    @In Identity identity;
    
    @In Credentials credentials;
	
    @In FacesMessages facesMessages;
    
    @In IUserAction userAction;
	
	private boolean loginFailed = false;

	private String digestAlgorithm;
	
	private String charset;

	public void setDigestAlgorithm(String algorithm) {
		this.digestAlgorithm = algorithm;
	}

	public void setCharset(String charset) {
		this.charset = charset;
	}

	public String hash(String plainTextPassword) {
		try {
			MessageDigest digest = MessageDigest.getInstance(digestAlgorithm);
			digest.update(plainTextPassword.getBytes(charset));
			byte[] rawHash = digest.digest();
			return new String(Hex.encodeHex(rawHash));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	public String login(){
		if(identity.hasRole(Constants.ADMIN_ROLE)){
			return null;
		}
		String password = credentials.getPassword();
		String username = credentials.getUsername();
		identity.unAuthenticate();
		credentials.setPassword(password);
		credentials.setUsername(username);
		try {
			identity.authenticate();		
		} catch (LoginException e) {
			setLoginFailed(true);
			credentials.clear();
			identity.login();
			identity.addRole(Constants.GUEST_ROLE);
			facesMessages.clear();
			facesMessages.add(Constants.INVALID_LOGIN_OR_PASSWORD);
			FacesContext.getCurrentInstance().renderResponse();
		}
		return Constants.MAIN_OUTCOME;
	}

	@End
    public String register(User user){
    	if(checkPassword(user) || checkUserExist(user)){
    		return "";
    	}
    	user.setPasswordHash(hash(user.getPassword()));
    	user.setPreDefined(false);
    	try{
    		userAction.register(user);
    	}
    	catch(Exception e){
    		return "";
    	}
    	return Constants.INDEX_OUTCOME;
    }
	
	@End
	public String goToIndex(){
		return Constants.INDEX_OUTCOME;
	}
	
	public String goToRegister(){
    	identity.unAuthenticate();
    	Contexts.getSessionContext().set("user", new User());
    	setLoginFailed(false);
    	return Constants.REGISTER_OUTCOME;
    }
	
	private boolean checkUserExist(User user) {
		if(userAction.isUserExist(user.getLogin())){
    		addFacesMessage(Constants.REGISTER_LOGIN_NAME_ID, Constants.USER_WITH_THIS_LOGIN_ALREADY_EXIST);
    		return true;
    	}
		return false;
	}

	private void addFacesMessage(String componentId, String message ) {
		UIComponent root = FacesContext.getCurrentInstance().getViewRoot();
		UIComponent component = root.findComponent(componentId);
		FacesContext.getCurrentInstance().addMessage(component.getClientId(FacesContext.getCurrentInstance()), new FacesMessage(
				FacesMessage.SEVERITY_ERROR, message , message));
	}

	private boolean checkPassword(User user) {
		if(!user.getPassword().equals(user.getConfirmPassword())){
			addFacesMessage(Constants.REGISTER_CONFIRM_PASSWORD_ID, Constants.CONFIRM_PASSWORD_NOT_EQUALS_PASSWORD);
			return true;
    	}
		return false;
	}
	
    public boolean authenticate()
    {
    	if (wantLoginAnonymous()) {
			identity.addRole(Constants.GUEST_ROLE);
			Events.instance().raiseEvent(Constants.UPDATE_MAIN_AREA_EVENT, NavigationEnum.ANONYM);
			return true;
		}
		try {
			User user = userAction.login(credentials.getUsername(), hash(credentials.getPassword()));
			if (user != null) {
				identity.addRole(Constants.ADMIN_ROLE);
				Events.instance().raiseEvent(Constants.AUTHENTICATED_EVENT, user);
				setLoginFailed(false);
				Contexts.getSessionContext().set(Constants.USER_VARIABLE, user);
				return true;
			}
		} catch (Exception nre) {
			return false;
		}
		return false; 
    }

    public String logout(){
    	identity.logout();
    	return Constants.LOGOUT_OUTCOME;
    }
    
    public String startConversation(){
    	if(!identity.isLoggedIn()){
        	credentials.clear();
        	try {
    			identity.authenticate();
    		} catch (LoginException e) {
    			//No cases for that
    		}
    	}
    	return "";
    }

    public String destroyConversation(){
    	if(identity.isLoggedIn()){
    		identity.logout();
    	}else{
    		identity.unAuthenticate();
    	}
		return "";
    }
    
	private boolean wantLoginAnonymous() {
		return null == credentials.getUsername() || credentials.getUsername().equals("");
	}

	public Long getStartConversation() {
		return startConversation;
	}

	public boolean isLoginFailed() {
		return loginFailed;
	}

	public void setLoginFailed(boolean loginFailed) {
		this.loginFailed = loginFailed;
	}
}